""" AtnActorBase """
import functools
import logging
import random
import time
import uuid
from abc import abstractmethod
from typing import Optional
from typing import no_type_check

import gridworks.algo_utils as algo_utils
import pendulum
from algosdk.atomic_transaction_composer import TransactionWithSigner
from algosdk.future.transaction import *
from algosdk.v2client.algod import AlgodClient
from beaker.client import ApplicationClient
from gridworks.algo_utils import BasicAccount
from gridworks.enums import GNodeRole
from gridworks.message import as_enum
from pydantic import BaseModel

from gwatn.config import AtnSettings
from gwatn.dispatch_contract import DispatchContract
from gwatn.enums import AlgoCertType
from gwatn.enums import MessageCategorySymbol
from gwatn.enums import UniverseType
from gwatn.two_channel_actor_base import TwoChannelActorBase
from gwatn.types import AtnParamsHeatpumpwithbooststore as AtnParams
from gwatn.types import (
    DispatchContractConfirmedHeatpumpwithbooststore_Maker as DispatchContractConfirmed_Maker,
)
from gwatn.types import GwCertId
from gwatn.types import GwCertId_Maker
from gwatn.types import HeartbeatA
from gwatn.types import HeartbeatA_Maker
from gwatn.types import HeartbeatB
from gwatn.types import HeartbeatB_Maker
from gwatn.types import JoinDispatchContract
from gwatn.types import JoinDispatchContract_Maker
from gwatn.types import LatestPrice
from gwatn.types import LatestPrice_Maker
from gwatn.types import SimTimestep
from gwatn.types import SimTimestep_Maker


LOG_FORMAT = (
    "%(levelname) -10s %(sasctime)s %(name) -30s %(funcName) "
    "-35s %(lineno) -5d: %(message)s"
)
LOGGER = logging.getLogger(__name__)


class HbStatus(BaseModel):
    LastHeartbeatReceivedMs: int
    AtnLastHex: str = "0"
    ScadaLastHex: str = "0"


def dummy_atn_params() -> AtnParams:
    return AtnParams(
        SliceDurationMinutes=60,
        FloSlices=48,
        GNodeAlias="d1.isone.dummy.ta",
        GNodeInstanceId="00000000-0000-0000-0000-000000000000",
        TypeName="atn.params.heatpumpwithbooststore",
        Version="000",
    )


class AtnActorBase(TwoChannelActorBase):
    def __init__(self, settings: AtnSettings):
        super().__init__(settings=settings)
        self.settings: AtnSettings = settings
        self.scada_gni_id = settings.scada_gni_id
        self.acct: BasicAccount = BasicAccount(settings.sk.get_secret_value())
        self.client: AlgodClient = AlgodClient(
            settings.algo_api_secrets.algod_token.get_secret_value(),
            settings.public.algod_address,
        )
        if algo_utils.algos(self.acct.addr) < 5:
            raise Exception(
                f"Insufficiently funded. Make sure atn has at least 5 algos"
            )
        # TODO: move this into spaceheat along with join_dispatch_contract_received
        self.atn_params: AtnParams = dummy_atn_params()
        self.sp = self.client.suggested_params()
        self.sp.flat_fee = True
        self.sp.fee = 2000
        # this is initialized with the AppId provided by the SCADA
        self.dc_app_id: Optional[int] = None
        self.dc_client: Optional[ApplicationClient] = None
        self.check_for_dispatch_contract()
        self.universe_type = as_enum(
            self.settings.universe_type_value, UniverseType, UniverseType.default()
        )
        self.trading_rights_id: Optional[GwCertId] = None
        self.update_trading_rights()
        self.hb_status = HbStatus(LastHeartbeatReceivedMs=int(time.time() * 1000))
        self._time: float = self.get_initial_time_s()

    def local_rabbit_startup(self) -> None:
        rjb = MessageCategorySymbol.rjb.value
        tc_alias_lrh = self.settings.time_coordinator_alias.replace(".", "-")
        binding = f"{rjb}.{tc_alias_lrh}.timecoordinator.sim-timestep"

        cb = functools.partial(self.on_timecoordinator_bindok, binding=binding)
        self._consume_channel.queue_bind(
            self.queue_name, "timecoordinatormic_tx", routing_key=binding, callback=cb
        )
        LOGGER.info(
            f"Queue {self.queue_name} bound to timecoordinatormic_tx with {binding} "
        )
        pong = HeartbeatA_Maker(
            my_hex=str(random.choice("0123456789abcdef")), your_last_hex="0"
        ).tuple
        self.send_message(
            payload=pong,
            to_role=GNodeRole.Supervisor,
            to_g_node_alias=self.settings.my_super_alias,
        )
        d = pendulum.from_timestamp(time.time())
        LOGGER.warning(
            f"[{self.alias}] Sent first heartbeat to super: {d.minute}:{d.second}.{d.microsecond}"
        )
        self.strategy_rabbit_startup()

    def strategy_rabbit_startup(self) -> None:
        pass

    @no_type_check
    def on_timecoordinator_bindok(self, _unused_frame, binding) -> None:
        LOGGER.info(f"Queue {self.queue_name} bound with {binding}")

    def time(self) -> float:
        if self.universe_type == UniverseType.Dev:
            return self._time
        else:
            return time.time()

    def get_initial_time_s(self) -> float:
        if self.universe_type == UniverseType.Dev:
            return self.settings.initial_time_unix_s
        else:
            return time.time()

    def prepare_for_death(self) -> None:
        self.actor_main_stopped = True

    ########################
    ## Receives
    ########################

    def route_message(
        self, from_alias: str, from_role: GNodeRole, payload: HeartbeatA
    ) -> None:
        self.payload = payload
        if payload.TypeName == HeartbeatA_Maker.type_name:
            if from_role != GNodeRole.Supervisor:
                LOGGER.info(
                    f"Ignoring HeartbeatA from GNode with role {from_role}; expects Supervisor"
                )
            try:
                self.heartbeat_from_super(from_alias, payload)
            except:
                LOGGER.exception("Error in heartbeat_received")
        elif payload.TypeName == HeartbeatB_Maker.type_name:
            if from_role != GNodeRole.Scada:
                LOGGER.info(
                    f"Ignoring HeartbeatB from GNode with role {from_role}; expects Scada"
                )
            try:
                self.heartbeat_from_scada(payload)
            except:
                LOGGER.exception("Error in heartbeat_from_partner")
        elif payload.TypeName == JoinDispatchContract_Maker.type_name:
            if from_role != GNodeRole.Scada:
                LOGGER.info(
                    f"Ignoring JoinDispatchContract from GNode with role {from_role}; expects Scada"
                )
            try:
                self.join_dispatch_contract_from_scada(payload)
            except:
                LOGGER.exception("join_dispatch_contract_from_scada")
        elif payload.TypeName == LatestPrice_Maker.type_name:
            if from_role == GNodeRole.MarketMaker:
                try:
                    self.latest_price_from_market_maker(payload)
                except:
                    LOGGER.exception("Error in latest_price_from_market_maker")
        elif payload.TypeName == SimTimestep_Maker.type_name:
            try:
                self.timestep_from_timecoordinator(payload)
            except:
                LOGGER.exception("Error in timestep_from_timecoordinator")

    def heartbeat_from_super(self, from_alias: str, ping: HeartbeatA) -> None:
        pong = HeartbeatA_Maker(
            my_hex=str(random.choice("0123456789abcdef")), your_last_hex=ping.MyHex
        ).tuple

        self.send_message(
            payload=pong,
            to_role=GNodeRole.Supervisor,
            to_g_node_alias=self.settings.my_super_alias,
        )

        LOGGER.debug(
            f"[{self.alias}] Sent HB: SuHex {pong.YourLastHex}, AtnHex {pong.MyHex}"
        )

    def heartbeat_from_scada(self, ping: HeartbeatB) -> None:
        """
        This is the Atn's half of the DispatchContract Heartbeat pattern.
        It:
          - Checks that it has opted into a DispatchContract
          - Checks the  FromGNodeInstanceId to validate Scada credentials
          - Checks that the scada correctly repeated last sent hex
          - Updates the last hex received (for use at the top of the next minute) along w time received
        [more info](https://gridworks.readthedocs.io/en/latest/dispatch-contract.html)

        Args:
            payload (HeartbeatB): The latest heartbeat received from its
            SCADA partner

        """
        if not self.in_dispatch_contract():
            LOGGER.info(f"Not in Dispatch Contract. Ignoring")
        if ping.FromGNodeInstanceId != self.scada_gni_id:
            raise Exception("In dispatch contract but mismatched Scada Gni Id!")
        if ping.YourLastHex != self.hb_status.AtnLastHex:
            LOGGER.info(f"Received incorrect pong. Ignoring")
            return
        self.hb_status.LastHeartbeatReceivedMs = int(time.time() * 1000)
        self.hb_status.ScadaLastHex = ping.MyHex
        self.hb_status.AtnLastHex = str(random.choice("0123456789abcdef"))
        LOGGER.info(f"Got {ping.MyHex}")
        # Does not send back. Atn waits for the DispatchContract's expected
        # one minute before sending.
        print(f"Got heartbeat from scada: {ping}")

    def hb_to_scada(self):
        """Checks that Atn is in Dispatch Contract, sends a HeartbeatB to Scada,
        and then reports to DispatchContract"""
        if not self.in_dispatch_contract():
            LOGGER.info("Not sending hb to Scada - not in DispatchContract yet")
        ping = HeartbeatB_Maker(
            from_g_node_alias=self.alias,
            from_g_node_instance_id=self.g_node_instance_id,
            my_hex=self.hb_status.AtnLastHex,
            your_last_hex=self.hb_status.ScadaLastHex,
            last_received_time_unix_ms=self.hb_status.LastHeartbeatReceivedMs,
            send_time_unix_ms=int(1000 * time.time()),
        ).tuple
        self.send_message(
            payload=ping,
            to_role=GNodeRole.Scada,
            to_g_node_alias=self.scada_alias,
        )
        LOGGER.info(f"Sent hb {ping}")
        # report to DispatchContract
        ptxn = PaymentTxn(self.acct.addr, self.sp, self.dc_client.app_addr, 1000)
        self.dc_client.call(
            DispatchContract.heartbeat_algo_audit,
            signed_proof=TransactionWithSigner(ptxn, self.acct.as_signer()),
            heartbeat=ping.as_dict(),
        )

    def check_for_dispatch_contract(self):
        """Upon startup, see if already party of a dispatch contract,
        and if so start up the dc_client and set the dc_app_id"""
        apps = self.client.account_info(self.acct.addr)["apps-local-state"]
        if len(apps) > 1:
            raise Exception("Should only be part of one app")
        if len(apps) == 0:
            return
        self.dc_app_id = apps[0]["id"]
        self.dc_client = ApplicationClient(
            client=self.client,
            app=DispatchContract(),
            signer=self.acct.as_signer(),
            app_id=self.dc_app_id,
        )

    def update_trading_rights(self) -> None:
        """Look at address to find our trading rights"""
        balances = algo_utils.get_balances(self.client, self.acct.addr)
        non_algo_asas = list(set(balances.keys()) - {0})
        ta_trading_rights = list(
            filter(
                lambda x: self.client.asset_info(x)["params"]["unit-name"] == "TATRADE",
                non_algo_asas,
            )
        )
        if len(ta_trading_rights) != 1:
            return
        x = ta_trading_rights[0]
        if self.client.asset_info(x)["params"]["name"] != self.ta_alias:
            raise Exception(
                f"Mismatch in trading rights and Alias. {self.alias} and {x}"
            )
        self.trading_rights_id = GwCertId_Maker(
            type=AlgoCertType.ASA, idx=ta_trading_rights[0], addr=None
        ).tuple

    def join_dispatch_contract_from_scada(self, payload: JoinDispatchContract) -> None:
        """Bootstraps the DispatchContract, opts in, and sends back DispatchContractConfirmed

        - ignores if already opted into a DispatchContract
        - Until World registry is in place, update g node instance id and write to .env
        - sets up ApplicationClient using the app_id sent by the scada
        - Checks that scada bootstrapping is done (contract has ta_alias and scada_cert_idx)
        - Check that our Atn acct owns the TaTradingRights for this TaAlias

        """
        if self.in_dispatch_contract():
            LOGGER.warning(
                f"Ignoring JoinDispatchContract - already in dispatch contract {self.dc_app_id}"
            )
            return
        self.g_node_instance_id = str(uuid.uuid4())
        self.scada_gni_id = payload.FromGNodeInstanceId
        with open(".env", "a") as file:
            file.write(f'ATN_G_NODE_INSTANCE_ID="{self.g_node_instance_id}"\n')
            file.write(f'ATN_SCADA_GNI_ID="{payload.FromGNodeInstanceId}"\n')
        self.dc_client = ApplicationClient(
            client=self.client,
            app=DispatchContract(),
            signer=self.acct.as_signer(),
            app_id=payload.DispatchContractAppId,
        )
        self.dc_app_id = payload.DispatchContractAppId
        app_state = self.dc_client.get_application_state()
        if not {"ta_alias", "scada_cert_idx"}.issubset(set(app_state.keys())):
            LOGGER.warning(
                f"AppId {payload.DispatchContractAppId} not done with scada bootstrap!"
            )
            return
        if not self.ta_alias == app_state["ta_alias"]:
            LOGGER.warning(
                f"Wrong TaAlias. Mine is {self.ta_alias}, DispatchContract has {app_state['ta_alias']}"
            )

        if self.trading_rights_id is None:
            self.update_trading_rights()
        if self.trading_rights_id is None:
            LOGGER.warning(f"Do not have trading rights! Cannot join dispatch contract")
            return
        ptxn = PaymentTxn(self.acct.addr, self.sp, self.dc_client.app_addr, 100_000)

        try:
            result = self.dc_client.call(
                DispatchContract.bootstrap2,
                atn_seed=TransactionWithSigner(ptxn, self.acct.as_signer()),
                TaTradingRights=self.trading_rights_id.Idx,
            )
        except Exception as e:
            LOGGER.warning(f"Failed to bootstrap DispatchContract: {e}")
            return

        # opts in
        self.dc_client.opt_in()
        LOGGER.info(f"Dispatch Contract {self.dc_app_id} is live")

        # sends back DispatchContractConfirmed

        txn = PaymentTxn(
            sender=self.acct.addr,
            receiver=self.dc_client.app_addr,
            sp=self.sp,
            amt=1000,
        )
        signed_txn = txn.sign(self.acct.sk)
        # don't need to actually send transaction for sig
        payload = DispatchContractConfirmed_Maker(
            from_g_node_alias=self.alias,
            from_g_node_instance_id=self.g_node_instance_id,
            signed_proof=encoding.msgpack_encode(signed_txn),
            atn_params=self.atn_params,
        ).tuple
        self.send_message(
            payload=payload, to_role=GNodeRole.Scada, to_g_node_alias=self.scada_alias
        )

    @abstractmethod
    def latest_price_from_market_maker(self, payload: LatestPrice) -> None:
        raise NotImplementedError

    def timestep_from_timecoordinator(self, payload: SimTimestep):
        if self._time == 0:
            self._time = payload.TimeUnixS
            self.new_timestep(payload)
            LOGGER.info(f"TIME STARTED: {self.time_str()}")
        elif self._time < payload.TimeUnixS:
            self._time = payload.TimeUnixS
            self.new_timestep(payload)
            LOGGER.debug(f"Time is now {self.time_str()}")
        elif self._time == payload.TimeUnixS:
            self.repeat_timestep(payload)

    @abstractmethod
    def new_timestep(self, payload: SimTimestep) -> None:
        # LOGGER.info("New timestep in atn_actor_base")
        raise NotImplementedError

    @abstractmethod
    def repeat_timestep(self, payload: SimTimestep) -> None:
        # LOGGER.info("Timestep received again in atn_actor_base")
        raise NotImplementedError

    def time_str(self) -> str:
        return pendulum.from_timestamp(self.time()).strftime("%m/%d/%Y, %H:%M")

    @property
    def ta_alias(self):
        """Add `ta` to the and of the Atn's GNodeAlias"""
        return self.alias + ".ta"

    @property
    def scada_alias(self):
        """Add `scada` from the end of the Atn's's GNodeAlias"""
        return self.alias + ".scada"

    def in_dispatch_contract(self) -> bool:
        """Checks that bootstrap 2 was completed w addition of ta_trading_rights_idx,
        and also that Scada is opted in"""
        if self.dc_client is None:
            return False
        app_state = self.dc_client.get_application_state()
        if "ta_trading_rights_idx" not in app_state.keys():
            return False
        # Todo: also check that atn is opted in
        apps = self.client.account_info(self.acct.addr)["apps-local-state"]
        app_ids = list(map(lambda x: x["id"], apps))
        if self.dc_client.app_id in app_ids:
            return True
        return False



from enum import auto
from typing import List

from fastapi_utils.enums import StrEnum


class CoreGNodeRole(StrEnum):
    """
    CoreGNodeRole assigned by GNodeFactory. [More Info](https://gridworks.readthedocs.io/en/latest/core-g-node-role.html).

    Choices and descriptions:

      * Other:
      * TerminalAsset:
      * AtomicTNode:
      * MarketMaker:
      * AtomicMeteringNode:
      * ConductorTopologyNode:
      * InterconnectionComponent:
      * Scada:
    """

    Other = auto()
    TerminalAsset = auto()
    AtomicTNode = auto()
    MarketMaker = auto()
    AtomicMeteringNode = auto()
    ConductorTopologyNode = auto()
    InterconnectionComponent = auto()
    Scada = auto()

    @classmethod
    def default(cls) -> "CoreGNodeRole":
        """
        Returns default value Other
        """
        return cls.Other

    @classmethod
    def values(cls) -> List[str]:
        """
        Returns enum choices
        """
        return [elt.value for elt in cls]



"""Type g.node.instance.gt, version 000"""
import json
from enum import auto
from typing import Any
from typing import Dict
from typing import List
from typing import Literal
from typing import Optional

from fastapi_utils.enums import StrEnum
from gridworks.errors import SchemaError
from gridworks.message import as_enum
from pydantic import BaseModel
from pydantic import Field
from pydantic import validator

from gwatn.data_classes.g_node_instance import GNodeInstance
from gwatn.enums import GniStatus
from gwatn.enums import StrategyName
from gwatn.types.g_node_gt import GNodeGt
from gwatn.types.g_node_gt import GNodeGt_Maker


class GniStatus000SchemaEnum:
    enum_name: str = "gni.status.000"
    symbols: List[str] = [
        "00000000",
        "7890ab0a",
        "69241259",
        "8222421f",
    ]

    @classmethod
    def is_symbol(cls, candidate: str) -> bool:
        if candidate in cls.symbols:
            return True
        return False


class GniStatus000(StrEnum):
    Unknown = auto()
    Pending = auto()
    Active = auto()
    Done = auto()

    @classmethod
    def default(cls) -> "GniStatus000":
        return cls.Unknown

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]


class GniStatusMap:
    @classmethod
    def type_to_local(cls, symbol: str) -> GniStatus:
        if not GniStatus000SchemaEnum.is_symbol(symbol):
            raise SchemaError(f"{symbol} must belong to GniStatus000 symbols")
        versioned_enum = cls.type_to_versioned_enum_dict[symbol]
        return as_enum(versioned_enum, GniStatus, GniStatus.default())

    @classmethod
    def local_to_type(cls, gni_status: GniStatus) -> str:
        if not isinstance(gni_status, GniStatus):
            raise SchemaError(f"{gni_status} must be of type {GniStatus}")
        versioned_enum = as_enum(gni_status, GniStatus000, GniStatus000.default())
        return cls.versioned_enum_to_type_dict[versioned_enum]

    type_to_versioned_enum_dict: Dict[str, GniStatus000] = {
        "00000000": GniStatus000.Unknown,
        "7890ab0a": GniStatus000.Pending,
        "69241259": GniStatus000.Active,
        "8222421f": GniStatus000.Done,
    }

    versioned_enum_to_type_dict: Dict[GniStatus000, str] = {
        GniStatus000.Unknown: "00000000",
        GniStatus000.Pending: "7890ab0a",
        GniStatus000.Active: "69241259",
        GniStatus000.Done: "8222421f",
    }


class StrategyName000SchemaEnum:
    enum_name: str = "strategy.name.000"
    symbols: List[str] = [
        "00000000",
        "642c83d3",
        "4bb2cf7e",
        "f5961401",
        "73fbe6ab",
        "5e18a52e",
    ]

    @classmethod
    def is_symbol(cls, candidate: str) -> bool:
        if candidate in cls.symbols:
            return True
        return False


class StrategyName000(StrEnum):
    NoActor = auto()
    WorldA = auto()
    SupervisorA = auto()
    AtnHeatPumpWithBoostStore = auto()
    TcGlobalA = auto()
    MarketMakerA = auto()

    @classmethod
    def default(cls) -> "StrategyName000":
        return cls.NoActor

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]


class StrategyNameMap:
    @classmethod
    def type_to_local(cls, symbol: str) -> StrategyName:
        if not StrategyName000SchemaEnum.is_symbol(symbol):
            raise SchemaError(f"{symbol} must belong to StrategyName000 symbols")
        versioned_enum = cls.type_to_versioned_enum_dict[symbol]
        return as_enum(versioned_enum, StrategyName, StrategyName.default())

    @classmethod
    def local_to_type(cls, strategy_name: StrategyName) -> str:
        if not isinstance(strategy_name, StrategyName):
            raise SchemaError(f"{strategy_name} must be of type {StrategyName}")
        versioned_enum = as_enum(
            strategy_name, StrategyName000, StrategyName000.default()
        )
        return cls.versioned_enum_to_type_dict[versioned_enum]

    type_to_versioned_enum_dict: Dict[str, StrategyName000] = {
        "00000000": StrategyName000.NoActor,
        "642c83d3": StrategyName000.WorldA,
        "4bb2cf7e": StrategyName000.SupervisorA,
        "f5961401": StrategyName000.AtnHeatPumpWithBoostStore,
        "73fbe6ab": StrategyName000.TcGlobalA,
        "5e18a52e": StrategyName000.MarketMakerA,
    }

    versioned_enum_to_type_dict: Dict[StrategyName000, str] = {
        StrategyName000.NoActor: "00000000",
        StrategyName000.WorldA: "642c83d3",
        StrategyName000.SupervisorA: "4bb2cf7e",
        StrategyName000.AtnHeatPumpWithBoostStore: "f5961401",
        StrategyName000.TcGlobalA: "73fbe6ab",
        StrategyName000.MarketMakerA: "5e18a52e",
    }


def check_is_reasonable_unix_time_s(v: int) -> None:
    """
    ReasonableUnixTimeS format: time in unix seconds between Jan 1 2000 and Jan 1 3000

    Raises:
        ValueError: if not ReasonableUnixTimeS format
    """
    import pendulum

    if pendulum.parse("2000-01-01T00:00:00Z").int_timestamp > v:  # type: ignore[attr-defined]
        raise ValueError(f"{v} must be after Jan 1 2000")
    if pendulum.parse("3000-01-01T00:00:00Z").int_timestamp < v:  # type: ignore[attr-defined]
        raise ValueError(f"{v} must be before Jan 1 3000")


def check_is_uuid_canonical_textual(v: str) -> None:
    """
    UuidCanonicalTextual format:  A string of hex words separated by hyphens
    of length 8-4-4-4-12.

    Raises:
        ValueError: if not UuidCanonicalTextual format
    """
    try:
        x = v.split("-")
    except AttributeError as e:
        raise ValueError(f"Failed to split on -: {e}")
    if len(x) != 5:
        raise ValueError(f"{v} split by '-' did not have 5 words")
    for hex_word in x:
        try:
            int(hex_word, 16)
        except ValueError:
            raise ValueError(f"Words of {v} are not all hex")
    if len(x[0]) != 8:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")
    if len(x[1]) != 4:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")
    if len(x[2]) != 4:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")
    if len(x[3]) != 4:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")
    if len(x[4]) != 12:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")


def check_is_algo_address_string_format(v: str) -> None:
    """
    AlgoAddressStringFormat format: The public key of a private/public Ed25519
    key pair, transformed into an  Algorand address, by adding a 4-byte checksum
    to the end of the public key and then encoding in base32.

    Raises:
        ValueError: if not AlgoAddressStringFormat format
    """
    import algosdk

    at = algosdk.abi.AddressType()
    try:
        result = at.decode(at.encode(v))
    except Exception as e:
        raise ValueError(f"Not AlgoAddressStringFormat: {e}")


class GNodeInstanceGt(BaseModel):
    """Used to send and receive updates about GNodeInstances.

    One of the layers of abstraction connecting a GNode with a running app in
    a Docker container.

    [More info](https://gridworks.readthedocs.io/en/latest/g-node-instance.html).
    """

    GNodeInstanceId: str = Field(
        title="Immutable identifier for GNodeInstance (Gni)",
    )
    GNode: GNodeGt = Field(
        title="The GNode represented by the Gni",
    )
    Strategy: StrategyName = Field(
        title="Used to determine the code running in a GNode actor application",
    )
    Status: GniStatus = Field(
        title="Lifecycle Status for Gni",
    )
    SupervisorContainerId: str = Field(
        title="The Id of the docker container where the Gni runs",
    )
    StartTimeUnixS: int = Field(
        title="When the gni starts representing the GNode",
        description="Specifically, when the Status changes from Pending to Active. Note that this is time in the GNode's World, which may not be real time if it is a simulation.",
    )
    EndTimeUnixS: int = Field(
        title="When the gni stops representing the GNode",
        description="Specifically, when the Status changes from Active to Done.",
    )
    AlgoAddress: Optional[str] = Field(
        title="Algorand address for Gni",
        default=None,
    )
    TypeName: Literal["g.node.instance.gt"] = "g.node.instance.gt"
    Version: str = "000"

    @validator("GNodeInstanceId")
    def _check_g_node_instance_id(cls, v: str) -> str:
        try:
            check_is_uuid_canonical_textual(v)
        except ValueError as e:
            raise ValueError(
                f"GNodeInstanceId failed UuidCanonicalTextual format validation: {e}"
            )
        return v

    @validator("Strategy")
    def _check_strategy(cls, v: StrategyName) -> StrategyName:
        return as_enum(v, StrategyName, StrategyName.NoActor)

    @validator("Status")
    def _check_status(cls, v: GniStatus) -> GniStatus:
        return as_enum(v, GniStatus, GniStatus.Unknown)

    @validator("SupervisorContainerId")
    def _check_supervisor_container_id(cls, v: str) -> str:
        try:
            check_is_uuid_canonical_textual(v)
        except ValueError as e:
            raise ValueError(
                f"SupervisorContainerId failed UuidCanonicalTextual format validation: {e}"
            )
        return v

    @validator("StartTimeUnixS")
    def _check_start_time_unix_s(cls, v: int) -> int:
        try:
            check_is_reasonable_unix_time_s(v)
        except ValueError as e:
            raise ValueError(
                f"StartTimeUnixS failed ReasonableUnixTimeS format validation: {e}"
            )
        return v

    @validator("AlgoAddress")
    def _check_algo_address(cls, v: Optional[str]) -> Optional[str]:
        if v is None:
            return v
        try:
            check_is_algo_address_string_format(v)
        except ValueError as e:
            raise ValueError(
                f"AlgoAddress failed AlgoAddressStringFormat format validation: {e}"
            )
        return v

    def as_dict(self) -> Dict[str, Any]:
        d = self.dict()
        d["GNode"] = self.GNode.as_dict()
        del d["Strategy"]
        Strategy = as_enum(self.Strategy, StrategyName, StrategyName.default())
        d["StrategyGtEnumSymbol"] = StrategyNameMap.local_to_type(Strategy)
        del d["Status"]
        Status = as_enum(self.Status, GniStatus, GniStatus.default())
        d["StatusGtEnumSymbol"] = GniStatusMap.local_to_type(Status)
        if d["AlgoAddress"] is None:
            del d["AlgoAddress"]
        return d

    def as_type(self) -> str:
        return json.dumps(self.as_dict())


class GNodeInstanceGt_Maker:
    type_name = "g.node.instance.gt"
    version = "000"

    def __init__(
        self,
        g_node_instance_id: str,
        g_node: GNodeGt,
        strategy: StrategyName,
        status: GniStatus,
        supervisor_container_id: str,
        start_time_unix_s: int,
        end_time_unix_s: int,
        algo_address: Optional[str],
    ):
        self.tuple = GNodeInstanceGt(
            GNodeInstanceId=g_node_instance_id,
            GNode=g_node,
            Strategy=strategy,
            Status=status,
            SupervisorContainerId=supervisor_container_id,
            StartTimeUnixS=start_time_unix_s,
            EndTimeUnixS=end_time_unix_s,
            AlgoAddress=algo_address,
            #
        )

    @classmethod
    def tuple_to_type(cls, tuple: GNodeInstanceGt) -> str:
        """
        Given a Python class object, returns the serialized JSON type object
        """
        return tuple.as_type()

    @classmethod
    def type_to_tuple(cls, t: str) -> GNodeInstanceGt:
        """
        Given a serialized JSON type object, returns the Python class object
        """
        try:
            d = json.loads(t)
        except TypeError:
            raise SchemaError("Type must be string or bytes!")
        if not isinstance(d, dict):
            raise SchemaError(f"Deserializing {t} must result in dict!")
        return cls.dict_to_tuple(d)

    @classmethod
    def dict_to_tuple(cls, d: dict[str, Any]) -> GNodeInstanceGt:
        d2 = dict(d)
        if "GNodeInstanceId" not in d2.keys():
            raise SchemaError(f"dict {d2} missing GNodeInstanceId")
        if "GNode" not in d2.keys():
            raise SchemaError(f"dict {d2} missing GNode")
        if not isinstance(d2["GNode"], dict):
            raise SchemaError(f"d['GNode'] {d2['GNode']} must be a GNodeGt!")
        g_node = GNodeGt_Maker.dict_to_tuple(d2["GNode"])
        d2["GNode"] = g_node
        if "StrategyGtEnumSymbol" not in d2.keys():
            raise SchemaError(f"dict {d2} missing StrategyGtEnumSymbol")
        if d2["StrategyGtEnumSymbol"] in StrategyName000SchemaEnum.symbols:
            d2["Strategy"] = StrategyNameMap.type_to_local(d2["StrategyGtEnumSymbol"])
        else:
            d2["Strategy"] = StrategyName.default()
        if "StatusGtEnumSymbol" not in d2.keys():
            raise SchemaError(f"dict {d2} missing StatusGtEnumSymbol")
        if d2["StatusGtEnumSymbol"] in GniStatus000SchemaEnum.symbols:
            d2["Status"] = GniStatusMap.type_to_local(d2["StatusGtEnumSymbol"])
        else:
            d2["Status"] = GniStatus.default()
        if "SupervisorContainerId" not in d2.keys():
            raise SchemaError(f"dict {d2} missing SupervisorContainerId")
        if "StartTimeUnixS" not in d2.keys():
            raise SchemaError(f"dict {d2} missing StartTimeUnixS")
        if "EndTimeUnixS" not in d2.keys():
            raise SchemaError(f"dict {d2} missing EndTimeUnixS")
        if "AlgoAddress" not in d2.keys():
            d2["AlgoAddress"] = None
        if "TypeName" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TypeName")

        return GNodeInstanceGt(
            GNodeInstanceId=d2["GNodeInstanceId"],
            GNode=d2["GNode"],
            Strategy=d2["Strategy"],
            Status=d2["Status"],
            SupervisorContainerId=d2["SupervisorContainerId"],
            StartTimeUnixS=d2["StartTimeUnixS"],
            EndTimeUnixS=d2["EndTimeUnixS"],
            AlgoAddress=d2["AlgoAddress"],
            TypeName=d2["TypeName"],
            Version="000",
        )

    @classmethod
    def tuple_to_dc(cls, t: GNodeInstanceGt) -> GNodeInstance:
        if t.GNodeInstanceId in GNodeInstance.by_id.keys():
            dc = GNodeInstance.by_id[t.GNodeInstanceId]
        else:
            dc = GNodeInstance(
                g_node_instance_id=t.GNodeInstanceId,
                g_node=GNodeGt_Maker.tuple_to_dc(t.GNode),
                strategy=t.Strategy,
                status=t.Status,
                supervisor_container_id=t.SupervisorContainerId,
                start_time_unix_s=t.StartTimeUnixS,
                end_time_unix_s=t.EndTimeUnixS,
                algo_address=t.AlgoAddress,
            )

        return dc

    @classmethod
    def dc_to_tuple(cls, dc: GNodeInstance) -> GNodeInstanceGt:
        t = GNodeInstanceGt_Maker(
            g_node_instance_id=dc.g_node_instance_id,
            g_node=GNodeGt_Maker.dc_to_tuple(dc.g_node),
            strategy=dc.strategy,
            status=dc.status,
            supervisor_container_id=dc.supervisor_container_id,
            start_time_unix_s=dc.start_time_unix_s,
            end_time_unix_s=dc.end_time_unix_s,
            algo_address=dc.algo_address,
        ).tuple
        return t

    @classmethod
    def type_to_dc(cls, t: str) -> GNodeInstance:
        return cls.tuple_to_dc(cls.type_to_tuple(t))

    @classmethod
    def dc_to_type(cls, dc: GNodeInstance) -> str:
        return cls.dc_to_tuple(dc).as_type()

    @classmethod
    def dict_to_dc(cls, d: dict[Any, str]) -> GNodeInstance:
        return cls.tuple_to_dc(cls.dict_to_tuple(d))



import logging
from typing import Optional

import dotenv
import gridworks.algo_utils as algo_utils
import gridworks.api_utils as api_utils
import gridworks.dev_utils.algo_setup as algo_setup
import requests
from algosdk import encoding
from algosdk.future import transaction
from algosdk.v2client.algod import AlgodClient
from gridworks.algo_utils import BasicAccount
from gridworks.algo_utils import MultisigAccount
from gridworks.utils import RestfulResponse
from rich.pretty import pprint

import gwatn.config as config
from gwatn.types import InitialTadeedAlgoCreate_Maker
from gwatn.types import InitialTadeedAlgoTransfer_Maker
from gwatn.types import TavalidatorcertAlgoCreate
from gwatn.types import TavalidatorcertAlgoCreate_Maker
from gwatn.types import TavalidatorcertAlgoTransfer
from gwatn.types import TavalidatorcertAlgoTransfer_Maker
from gwatn.types import TerminalassetCertifyHack


LOGGER = logging.getLogger(__name__)


class DevValidator:
    def __init__(
        self,
        settings: config.ValidatorSettings = config.ValidatorSettings(
            _env_file=dotenv.find_dotenv()
        ),
    ):
        self.settings = settings
        self.acct: BasicAccount = BasicAccount(
            private_key=self.settings.sk.get_secret_value()
        )
        self.client: AlgodClient = AlgodClient(
            settings.algo_api_secrets.algod_token.get_secret_value(),
            settings.public.algod_address,
        )
        self.validator_multi: MultisigAccount = (
            api_utils.get_validator_account_with_admin(self.acct.addr)
        )
        algo_setup.dev_fund_to_min(
            addr=self.acct.addr,
            min_algos=self.settings.public.ta_validator_funding_threshold_algos + 1,
        )
        LOGGER.info(
            f"mollyMetermaid acct {self.acct.addr_short_hand} balance: {algo_utils.algos(self.acct.addr)} Algos"
        )

        self.seed_fund_validator_joint_account()
        LOGGER.info("DevValidator Initialized")

    #################
    # Messages received
    ################

    def terminalasset_certify_hack_received(
        self, payload: TerminalassetCertifyHack
    ) -> RestfulResponse:
        ta_alias = payload.TerminalAssetAlias
        r = self.post_initial_tadeed_algo_create(ta_alias=ta_alias)
        return r

    ###################
    # Messages sent
    ###################

    def post_initial_tadeed_algo_create(self, ta_alias: str) -> RestfulResponse:
        txn = transaction.AssetCreateTxn(
            sender=self.validator_multi.address(),
            total=1,
            decimals=0,
            default_frozen=False,
            manager=self.settings.public.gnf_admin_addr,
            asset_name=ta_alias,
            unit_name="TADEED",
            sp=self.client.suggested_params(),
        )
        mtx = self.validator_multi.create_mtx(txn)
        mtx.sign(self.acct.sk)

        payload = InitialTadeedAlgoCreate_Maker(
            validator_addr=self.acct.addr,
            half_signed_deed_creation_mtx=encoding.msgpack_encode(mtx),
        ).tuple

        api_endpoint = (
            f"{self.settings.public.gnf_api_root}/initial-tadeed-algo-create/"
        )
        request_response = requests.post(url=api_endpoint, json=payload.as_dict())

        if request_response.status_code > 200:
            if "detail" in request_response.json().keys():
                note = request_response.json()["detail"]
            else:
                note = request_response.reason
            r = RestfulResponse(Note=note, HttpStatusCode=422)
            return r
        r = RestfulResponse(**request_response.json())
        return r

    def post_tavalidatorcert_algo_create(self) -> RestfulResponse:
        txn = transaction.AssetCreateTxn(
            sender=self.validator_multi.address(),
            total=1,
            decimals=0,
            default_frozen=False,
            manager=self.settings.public.gnf_admin_addr,
            asset_name=self.settings.cert_name,
            unit_name="VLDTR",
            note=self.settings.name,
            sp=self.client.suggested_params(),
        )

        mtx = self.validator_multi.create_mtx(txn)
        mtx.sign(self.acct.sk)

        payload = TavalidatorcertAlgoCreate(
            ValidatorAddr=self.acct.addr,
            HalfSignedCertCreationMtx=encoding.msgpack_encode(mtx),
        )
        LOGGER.info("Posting request to GnfRestAPI to create new TaValidatorCert")
        api_endpoint = (
            f"{self.settings.public.gnf_api_root}/tavalidatorcert-algo-create/"
        )

        r = requests.post(url=api_endpoint, json=payload.as_dict())

        if r.status_code > 200:
            LOGGER.warning(r.json())
            if "detail" in r.json().keys():
                note = "TavalidatorcertAlgoCreate error:" + r.json()["detail"]
            else:
                note = r.reason
            return RestfulResponse(Note=note, HttpStatusCode=422)

        rr = RestfulResponse(**r.json())
        cert_idx = rr.PayloadAsDict["Value"]

        payload = self.generate_transfer_tavalidatorcert_algo(cert_idx=cert_idx)

        api_endpoint = (
            f"{self.settings.public.gnf_api_root}/tavalidatorcert-algo-transfer/"
        )
        r = requests.post(url=api_endpoint, json=payload.as_dict())

        if r.status_code > 200:
            if r.status_code == 422:
                note = "TavalidatorcertAlgoTransfer error:" + r.json()["detail"]
            else:
                note = r.reason
            rr = RestfulResponse(Note=note, HttpStatusCode=422)
            return rr

        return RestfulResponse(**r.json())

    def certify_terminal_asset(
        self,
        ta_deed_idx: int,
        ta_daemon_addr: str,
        ta_owner_addr: str,
        micro_lat: int,
        micro_lon: int,
    ) -> RestfulResponse:
        """
        This method is supposed to be called exactly for the FIRST time a TaDeed
        NFT is created for this ta_owner. For updated deeds, uses ExchangeTadeedAlgo

          - Creates the InitialTadeedAlgoTransfer payload and sends it to the Gnf
          - Returns the payload
        Args:
            ta_deed_idx (int): asset id of the TaDeed NFT
            ta_daemon (BasicAccount): The Layer 1 contract supporting NFT ownership
            and creation (TaDeed, TaTradingRights)

        Returns:
            InitialTadeedAlgoTransfer if sent,
            None if ta_multi does not have ta_deed_consideration_algos
        """

        required_algos = config.Public().ta_deed_consideration_algos
        if algo_utils.algos(ta_daemon_addr) < required_algos:
            note = (
                "ta_daemon_addr not sufficiently funded! Has "
                f"{algo_utils.algos(ta_daemon_addr)} Algos and needs "
                f"{required_algos} Algos"
            )
            return RestfulResponse(
                Note=note,
                HttpStatusCode=422,
            )
        txn = transaction.AssetTransferTxn(
            sender=self.validator_multi.addr,
            receiver=ta_daemon_addr,
            amt=1,
            index=ta_deed_idx,
            sp=self.client.suggested_params(),
        )
        mtx = self.validator_multi.create_mtx(txn)
        mtx.sign(self.acct.sk)
        payload = InitialTadeedAlgoTransfer_Maker(
            micro_lat=micro_lat,
            micro_lon=micro_lon,
            validator_addr=self.acct.addr,
            ta_daemon_addr=ta_daemon_addr,
            ta_owner_addr=ta_owner_addr,
            first_deed_transfer_mtx=encoding.msgpack_encode(mtx),
        ).tuple

        api_endpoint = (
            f"{self.settings.public.gnf_api_root}/initial-tadeed-algo-transfer/"
        )
        LOGGER.info(f"Using api_endpoint of {api_endpoint}")
        try:
            r = requests.post(url=api_endpoint, json=payload.as_dict())
        except Exception as e:
            raise Exception(f"DevValidator failed to post request! {e}")

        LOGGER.info("Response from GnfRestAPI:")

        if r.status_code > 200:
            if r.status_code == 422:
                note = "TavalidatorcertAlgoTransfer error:" + r.json()["detail"]
            else:
                note = r.reason
            r = RestfulResponse(Note=note, HttpStatusCode=r.status_code)
            return r

        pprint(r.json())
        r = RestfulResponse(**r.json())
        return r

    def generate_transfer_tavalidatorcert_algo(
        self, cert_idx: int
    ) -> TavalidatorcertAlgoTransfer:
        """First, opts in to the validator cert asset. Then, generates and signs the
        multsig transaction for transfer from the multi account self.multi (joint w
        gnf, threshold 2). Creates the TavalidatorcertAlgoTransferpayload with this
        mtx and sends it to the gnf.

        Args:
            cert_idx (int): the asset index for this validator's cert

        Returns:
            TavalidatorcertAlgoTransfer: the payload sent to the Gnf.
        """

        # Opting in to the cert
        opt_in_txn = transaction.AssetOptInTxn(
            sender=self.acct.addr,
            index=cert_idx,
            sp=self.client.suggested_params(),
        )
        signed_txn = opt_in_txn.sign(self.acct.sk)
        self.client.send_transaction(signed_txn)
        LOGGER.info(f"Molly has opted into asset idx {cert_idx}")
        algo_utils.wait_for_transaction(self.client, signed_txn.get_txid())

        transfer_txn = transaction.AssetTransferTxn(
            sender=self.validator_multi.addr,
            receiver=self.acct.addr,
            amt=1,
            index=cert_idx,
            sp=self.client.suggested_params(),
        )

        mtx = self.validator_multi.create_mtx(transfer_txn)
        mtx.sign(self.acct.sk)

        payload = TavalidatorcertAlgoTransfer(
            ValidatorAddr=self.acct.addr,
            HalfSignedCertTransferMtx=encoding.msgpack_encode(mtx),
        )
        return payload

    ##########################
    # dev methods
    ########################

    def seed_fund_validator_joint_account(
        self,
    ) -> Optional[algo_utils.PendingTxnResponse]:
        """Becoming a validator requires this"""
        required_algos = self.settings.public.ta_validator_funding_threshold_algos
        current_algos = algo_utils.algos(self.validator_multi.address())
        if current_algos >= required_algos:
            LOGGER.info(
                f"ValidatorMulti account already has {current_algos} Algos. No more funding required"
            )
            return
        txn_response = algo_utils.pay_account(
            client=self.client,
            sender=self.acct,
            to_addr=self.validator_multi.address(),
            amt_in_micros=required_algos * 10**6,
        )
        if algo_utils.algos(self.validator_multi.address()) < required_algos:
            raise Exception(
                f"Failed to seed fund validator account {self.validator_multi.address()[-6:]}"
            )
        return txn_response



from enum import auto
from typing import List

from fastapi_utils.enums import StrEnum


class RecognizedCurrencyUnit(StrEnum):
    """
    Unit of currency

    Choices and descriptions:

      * Unknown:
      * USD: US Dollar
      * GBP: Pounds sterling
    """

    UNKNOWN = auto()
    USD = auto()
    GBP = auto()

    @classmethod
    def default(cls) -> "RecognizedCurrencyUnit":
        """
        Returns default value UNKNOWN
        """
        return cls.UNKNOWN

    @classmethod
    def values(cls) -> List[str]:
        """
        Returns enum choices
        """
        return [elt.value for elt in cls]



from typing import List
from typing import Optional

import gridworks.conversion_factors as cf
import numpy as np
import pendulum
from pydantic import BaseModel

import gwatn.errors as errors
import gwatn.property_format as property_format
import gwatn.utils as utils
from gwatn.data_classes import MarketType
from gwatn.enums import EmitterPumpFeedbackModel
from gwatn.enums import MixingValveFeedbackModel
from gwatn.schemata import AtnBid
from gwatn.schemata import AtnParamsHeatpumpwithbooststore as AtnParams
from gwatn.schemata import FloParamsHeatpumpwithbooststore as FloParams
from gwatn.schemata import FloParamsHeatpumpwithbooststore_Maker as FloParams_Maker
from gwatn.schemata import MarketSlot
from gwatn.schemata import MarketTypeGt_Maker
from gwatn.schemata import PriceQuantityUnitless
from gwatn.strategies.heatpumpwithbooststore.flo import (
    HeatPumpWithBoostStore__Flo as Flo,
)


DUMMY_ALGO_TXN = "gqNzaWfEQNPXbrAiWd+cNgsIaM3N0PSu3repauvmjuHmoKjh6sd3L5U4/YpovcXN7/ATH1LgcI4cgV+SU3VQ6bsm/gfAOQyjdHhuiaNhbXTNB9CjZmVlzQPoomZ2HaNnZW6qc2FuZG5ldC12MaJnaMQgaJXPYTdWaeTNSs8FMMzPNfV7SrHXqJgFsJLxRbSPjzCibHbNBAWjcmN2xCBWkH3PValty0Rb0cyZo69Alhp4IbNKFnhXtgJ++A9EzKNzbmTEIOJPEbccL6IqmeBeaLzbLav25U9jBMjloaIyF1eY9HFxpHR5cGWjcGF5"


class DijkstraChoice(BaseModel):
    PowerImportedKw: float
    NextNodeCostDollars: float


class CostAndQuantityBought(BaseModel):
    QuantityBought: float
    Cost: float


class SlotStuff(BaseModel):
    Slot: MarketSlot
    BidParams: Optional[FloParams] = None
    Flo: Optional[Flo] = None
    Bid: Optional[AtnBid] = None
    Price: Optional[float] = None


def dummy_bid() -> AtnBid:
    return AtnBid(
        BidderAlias="d1.isone.dummy.ta",
        BidderGNodeInstanceId="00000000-0000-0000-0000-000000000000",
        MarketSlotName="rt60gate30b.d1.isone.ver.keene.1577836800",
        BidList=[],
        SignedMarketFeeTxn=DUMMY_ALGO_TXN,
    )


def dummy_slot_stuff(slot: MarketSlot) -> SlotStuff:
    return SlotStuff(
        Slot=slot, FloStartIdx=50, BidParams=dummy_flo_params(), AtnBid=dummy_bid()
    )


def is_dummy_slot_stuff(bid_stuff: SlotStuff) -> bool:
    if is_dummy_flo_params(bid_stuff.BidParams):
        return True
    return False


def name_from_market_slot(slot: MarketSlot) -> str:
    return f"{slot.Type.Name.value}.{slot.MarketMakerAlias}.{slot.StartUnixS}"


def market_slot_from_name(market_slot_name: str) -> MarketSlot:
    """rt60gate30b.d1.isone.ver.keene.1577836800"""
    if not property_format.is_market_slot_name_lrd_format(market_slot_name):
        raise Exception(
            f"market slot alias {market_slot_name} does not have market"
            " slot alias lrd format!"
        )
    words = market_slot_name.split(".")
    market_type_name = words[0]
    market_type_dc = MarketType.by_id[market_type_name]
    market_type = MarketTypeGt_Maker.dc_to_tuple(market_type_dc)
    market_maker_alias = ".".join(words[1:-1])
    slot_start = int(words[-1])
    return MarketSlot(
        Type=market_type, MarketMakerAlias=market_maker_alias, StartUnixS=slot_start
    )


def is_dummy_atn_params(atn_params: AtnParams) -> bool:
    if atn_params.GNodeAlias == "d1.isone.dummy.ta":
        return True
    return False


def is_dummy_flo_params(flo_params: FloParams) -> bool:
    if flo_params.RtElecPriceUid == "00000000-0000-0000-0000-000000000000":
        return True
    return False


def dummy_flo_params() -> FloParams:
    return FloParams(
        RtElecPriceUid="00000000-0000-0000-0000-000000000000",
        WeatherUid="00000000-0000-0000-0000-000000000000",
    )


def get_k(atn_params: AtnParams) -> float:
    dt = atn_params.SystemMaxHeatOutputDeltaTempF
    gpm = atn_params.SystemMaxHeatOutputGpm
    swt = atn_params.SystemMaxHeatOutputSwtF
    rt = atn_params.RoomTempF
    return float(gpm * np.log(1 - dt / (swt - rt)))


def get_system_max_heat_output_kw_avg(atn_params: AtnParams) -> float:
    """What is the max heat that the system put out?  ASSUMES that the hydronic
    fluid is water.

    Args:
        tea_params: Uses the system_max_heat_output_delta_temp_f and
        the system_max_heat_output_gpm.

    Returns:
        float: max heat ouput of the heating system, in kw
    """
    gpm = atn_params.SystemMaxHeatOutputGpm
    delta_temp = atn_params.SystemMaxHeatOutputDeltaTempF
    c = cf.POUNDS_OF_WATER_PER_GALLON * cf.MINUTES_PER_HOUR / cf.BTU_PER_KWH
    if gpm == 0 or delta_temp == 0:
        raise Exception(
            f"max gpm is {gpm}, max delta_temp is {delta_temp}. Cannot calculate system heat output."
        )
    pwr = c * delta_temp * gpm
    return pwr


def get_max_store_kwh_th(params: AtnParams) -> float:
    return (
        cf.KWH_TH_PER_GALLON_PER_DEG_F
        * params.StoreSizeGallons
        * (params.MaxStoreTempF - params.ZeroPotentialEnergyWaterTempF)
    )



from enum import auto
from typing import List

from fastapi_utils.enums import StrEnum


class SupervisorContainerStatus(StrEnum):
    """
    Manages lifecycle of the docker containers where GridWorks actors run

    Choices and descriptions:

      * Unknown: Default value
      * Authorized: World has created the information for starting the container
      * Launching: World has launched the container
      * Provisioning: Container has started, but is going through its provisioning process
      * Running: GNode actors in the container are active
      * Stopped: Stopped
      * Deleted: Deleted
    """

    Unknown = auto()
    Authorized = auto()
    Launching = auto()
    Provisioning = auto()
    Running = auto()
    Stopped = auto()
    Deleted = auto()

    @classmethod
    def default(cls) -> "SupervisorContainerStatus":
        """
        Returns default value Unknown
        """
        return cls.Unknown

    @classmethod
    def values(cls) -> List[str]:
        """
        Returns enum choices
        """
        return [elt.value for elt in cls]



"""Type market.slot, version 000"""
import json
from typing import Any
from typing import Dict
from typing import Literal

from gridworks.errors import SchemaError
from pydantic import BaseModel
from pydantic import Field
from pydantic import validator

from gwatn.types.market_type_gt import MarketTypeGt
from gwatn.types.market_type_gt import MarketTypeGt_Maker


def check_is_reasonable_unix_time_s(v: int) -> None:
    """
    ReasonableUnixTimeS format: time in unix seconds between Jan 1 2000 and Jan 1 3000

    Raises:
        ValueError: if not ReasonableUnixTimeS format
    """
    import pendulum

    if pendulum.parse("2000-01-01T00:00:00Z").int_timestamp > v:  # type: ignore[attr-defined]
        raise ValueError(f"{v} must be after Jan 1 2000")
    if pendulum.parse("3000-01-01T00:00:00Z").int_timestamp < v:  # type: ignore[attr-defined]
        raise ValueError(f"{v} must be before Jan 1 3000")


def check_is_left_right_dot(v: str) -> None:
    """
    LeftRightDot format: Lowercase alphanumeric words separated by periods,
    most significant word (on the left) starting with an alphabet character.

    Raises:
        ValueError: if not LeftRightDot format
    """
    from typing import List

    try:
        x: List[str] = v.split(".")
    except:
        raise ValueError(f"Failed to seperate {v} into words with split'.'")
    first_word = x[0]
    first_char = first_word[0]
    if not first_char.isalpha():
        raise ValueError(f"Most significant word of {v} must start with alphabet char.")
    for word in x:
        if not word.isalnum():
            raise ValueError(f"words of {v} split by by '.' must be alphanumeric.")
    if not v.islower():
        raise ValueError(f"All characters of {v} must be lowercase.")


class MarketSlot(BaseModel):
    """MarketSlot
    [More info](https://gridworks.readthedocs.io/en/latest/market-slot.html).
    """

    Type: MarketTypeGt = Field(
        title="Type",
    )
    MarketMakerAlias: str = Field(
        title="MarketMakerAlias",
    )
    StartUnixS: int = Field(
        title="StartUnixS",
    )
    TypeName: Literal["market.slot"] = "market.slot"
    Version: str = "000"

    @validator("MarketMakerAlias")
    def _check_market_maker_alias(cls, v: str) -> str:
        try:
            check_is_left_right_dot(v)
        except ValueError as e:
            raise ValueError(
                f"MarketMakerAlias failed LeftRightDot format validation: {e}"
            )
        return v

    @validator("StartUnixS")
    def _check_start_unix_s(cls, v: int) -> int:
        try:
            check_is_reasonable_unix_time_s(v)
        except ValueError as e:
            raise ValueError(
                f"StartUnixS failed ReasonableUnixTimeS format validation: {e}"
            )
        return v

    def as_dict(self) -> Dict[str, Any]:
        d = self.dict()
        d["Type"] = self.Type.as_dict()
        return d

    def as_type(self) -> str:
        return json.dumps(self.as_dict())


class MarketSlot_Maker:
    type_name = "market.slot"
    version = "000"

    def __init__(self, type: MarketTypeGt, market_maker_alias: str, start_unix_s: int):
        self.tuple = MarketSlot(
            Type=type,
            MarketMakerAlias=market_maker_alias,
            StartUnixS=start_unix_s,
            #
        )

    @classmethod
    def tuple_to_type(cls, tuple: MarketSlot) -> str:
        """
        Given a Python class object, returns the serialized JSON type object
        """
        return tuple.as_type()

    @classmethod
    def type_to_tuple(cls, t: str) -> MarketSlot:
        """
        Given a serialized JSON type object, returns the Python class object
        """
        try:
            d = json.loads(t)
        except TypeError:
            raise SchemaError("Type must be string or bytes!")
        if not isinstance(d, dict):
            raise SchemaError(f"Deserializing {t} must result in dict!")
        return cls.dict_to_tuple(d)

    @classmethod
    def dict_to_tuple(cls, d: dict[str, Any]) -> MarketSlot:
        d2 = dict(d)
        if "Type" not in d2.keys():
            raise SchemaError(f"dict {d2} missing Type")
        if not isinstance(d2["Type"], dict):
            raise SchemaError(f"d['Type'] {d2['Type']} must be a MarketTypeGt!")
        type = MarketTypeGt_Maker.dict_to_tuple(d2["Type"])
        d2["Type"] = type
        if "MarketMakerAlias" not in d2.keys():
            raise SchemaError(f"dict {d2} missing MarketMakerAlias")
        if "StartUnixS" not in d2.keys():
            raise SchemaError(f"dict {d2} missing StartUnixS")
        if "TypeName" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TypeName")

        return MarketSlot(
            Type=d2["Type"],
            MarketMakerAlias=d2["MarketMakerAlias"],
            StartUnixS=d2["StartUnixS"],
            TypeName=d2["TypeName"],
            Version="000",
        )



from enum import auto
from typing import List

from fastapi_utils.enums import StrEnum


class RecognizedTemperatureUnit(StrEnum):
    """
    Unit of temperature

    Choices and descriptions:

      * C: Celcius
      * F: Fahrenheit
    """

    C = auto()
    F = auto()

    @classmethod
    def default(cls) -> "RecognizedTemperatureUnit":
        """
        Returns default value C
        """
        return cls.C

    @classmethod
    def values(cls) -> List[str]:
        """
        Returns enum choices
        """
        return [elt.value for elt in cls]



from enum import auto
from typing import List

from fastapi_utils.enums import StrEnum


class MarketQuantityUnit(StrEnum):
    """
    Quantity unit assigned to MarketMaker MarketType

    Choices and descriptions:

      * AvgMW:
      * AvgkW:
    """

    AvgMW = auto()
    AvgkW = auto()

    @classmethod
    def default(cls) -> "MarketQuantityUnit":
        """
        Returns default value AvgMW
        """
        return cls.AvgMW

    @classmethod
    def values(cls) -> List[str]:
        """
        Returns enum choices
        """
        return [elt.value for elt in cls]



from enum import auto
from typing import List

from fastapi_utils.enums import StrEnum


class GNodeStatus(StrEnum):
    """
    Enum for managing GNode lifecycle. [More Info](https://gridworks.readthedocs.io/en/latest/g-node-status.html).

    Choices and descriptions:

      * Unknown: Default value
      * Pending: The GNode exists but cannot be used yet.
      * Active: The GNode can be used.
      * PermanentlyDeactivated: The GNode can no longer be used, now or in the future.
      * Suspended: The GNode cannot be used, but may become active in the future.
    """

    Unknown = auto()
    Pending = auto()
    Active = auto()
    PermanentlyDeactivated = auto()
    Suspended = auto()

    @classmethod
    def default(cls) -> "GNodeStatus":
        """
        Returns default value Unknown
        """
        return cls.Unknown

    @classmethod
    def values(cls) -> List[str]:
        """
        Returns enum choices
        """
        return [elt.value for elt in cls]



"""Type latest.price, version 000"""
import json
from enum import auto
from typing import Any
from typing import Dict
from typing import List
from typing import Literal
from typing import Optional

from fastapi_utils.enums import StrEnum
from gridworks.errors import SchemaError
from gridworks.message import as_enum
from pydantic import BaseModel
from pydantic import Field
from pydantic import validator

from gwatn.enums import MarketPriceUnit


class MarketPriceUnit000SchemaEnum:
    enum_name: str = "market.price.unit.000"
    symbols: List[str] = [
        "00000000",
    ]

    @classmethod
    def is_symbol(cls, candidate: str) -> bool:
        if candidate in cls.symbols:
            return True
        return False


class MarketPriceUnit000(StrEnum):
    USDPerMWh = auto()

    @classmethod
    def default(cls) -> "MarketPriceUnit000":
        return cls.USDPerMWh

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]


class MarketPriceUnitMap:
    @classmethod
    def type_to_local(cls, symbol: str) -> MarketPriceUnit:
        if not MarketPriceUnit000SchemaEnum.is_symbol(symbol):
            raise SchemaError(f"{symbol} must belong to MarketPriceUnit000 symbols")
        versioned_enum = cls.type_to_versioned_enum_dict[symbol]
        return as_enum(versioned_enum, MarketPriceUnit, MarketPriceUnit.default())

    @classmethod
    def local_to_type(cls, market_price_unit: MarketPriceUnit) -> str:
        if not isinstance(market_price_unit, MarketPriceUnit):
            raise SchemaError(f"{market_price_unit} must be of type {MarketPriceUnit}")
        versioned_enum = as_enum(
            market_price_unit, MarketPriceUnit000, MarketPriceUnit000.default()
        )
        return cls.versioned_enum_to_type_dict[versioned_enum]

    type_to_versioned_enum_dict: Dict[str, MarketPriceUnit000] = {
        "00000000": MarketPriceUnit000.USDPerMWh,
    }

    versioned_enum_to_type_dict: Dict[MarketPriceUnit000, str] = {
        MarketPriceUnit000.USDPerMWh: "00000000",
    }


def check_is_uuid_canonical_textual(v: str) -> None:
    """
    UuidCanonicalTextual format:  A string of hex words separated by hyphens
    of length 8-4-4-4-12.

    Raises:
        ValueError: if not UuidCanonicalTextual format
    """
    try:
        x = v.split("-")
    except AttributeError as e:
        raise ValueError(f"Failed to split on -: {e}")
    if len(x) != 5:
        raise ValueError(f"{v} split by '-' did not have 5 words")
    for hex_word in x:
        try:
            int(hex_word, 16)
        except ValueError:
            raise ValueError(f"Words of {v} are not all hex")
    if len(x[0]) != 8:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")
    if len(x[1]) != 4:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")
    if len(x[2]) != 4:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")
    if len(x[3]) != 4:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")
    if len(x[4]) != 12:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")


def check_is_iso_format(v: str) -> None:
    import datetime

    try:
        datetime.datetime.fromisoformat(v.replace("Z", "+00:00"))
    except:
        raise ValueError(f"{v} is not IsoFormat")


def check_is_left_right_dot(v: str) -> None:
    """
    LeftRightDot format: Lowercase alphanumeric words separated by periods,
    most significant word (on the left) starting with an alphabet character.

    Raises:
        ValueError: if not LeftRightDot format
    """
    from typing import List

    try:
        x: List[str] = v.split(".")
    except:
        raise ValueError(f"Failed to seperate {v} into words with split'.'")
    first_word = x[0]
    first_char = first_word[0]
    if not first_char.isalpha():
        raise ValueError(f"Most significant word of {v} must start with alphabet char.")
    for word in x:
        if not word.isalnum():
            raise ValueError(f"words of {v} split by by '.' must be alphanumeric.")
    if not v.islower():
        raise ValueError(f"All characters of {v} must be lowercase.")


def check_is_market_type_name_lrd_format(v: str) -> None:
    from gwatn.enums import MarketTypeName

    try:
        x = v.split(".")
    except AttributeError:
        raise ValueError(f"{v} failed to split on '.'")
    if not x[0] in MarketTypeName.values():
        raise ValueError(f"{v} not recognized MarketType")
    g_node_alias = ".".join(x[1:])
    check_is_left_right_dot(g_node_alias)


def check_is_market_slot_name_lrd_format(v: str) -> None:
    """
    MaketSlotNameLrdFormat: the format of a MarketSlotName.
      - The first word must be a MarketTypeName
      - The last word (unix time of market slot start) must
      be a 10-digit integer divisible by 300 (i.e. all MarketSlots
      start at the top of 5 minutes)
      - More strictly, the last word must be the start of a
      MarketSlot for that MarketType (i.e. divisible by 3600
      for hourly markets)
      - The middle words have LeftRightDot format (GNodeAlias
      of the MarketMaker)
    Example: rt60gate5.d1.isone.ver.keene.1673539200

    """
    from gwatn.data_classes.market_type import MarketType

    try:
        x = v.split(".")
    except AttributeError:
        raise ValueError(f"{v} failed to split on '.'")
    slot_start = x[-1]
    if len(slot_start) != 10:
        raise ValueError(f"slot start {slot_start} not of length 10")
    try:
        slot_start = int(slot_start)
    except ValueError:
        raise ValueError(f"slot start {slot_start} not an int")
    if slot_start % 300 != 0:
        raise ValueError(f"slot start {slot_start} not a multiple of 300")

    market_type_name_lrd = ".".join(x[:-1])
    try:
        check_is_market_type_name_lrd_format(market_type_name_lrd)
    except ValueError as e:
        raise ValueError(f"e")

    market_type = MarketType.by_id[market_type_name_lrd.split(".")[0]]
    if not slot_start % (market_type.duration_minutes * 60) == 0:
        raise ValueError(
            f"market_slot_start_s mod {(market_type.duration_minutes * 60)} must be 0"
        )


class LatestPrice(BaseModel):
    """Latest Price for a MarketType, sent by a MarketMaker.

    The price of the current MarketSlot
    [More info](https://gridworks.readthedocs.io/en/latest/market-slot.html).
    """

    FromGNodeAlias: str = Field(
        title="FromGNodeAlias",
    )
    FromGNodeInstanceId: str = Field(
        title="FromGNodeInstanceId",
    )
    PriceTimes1000: int = Field(
        title="PriceTimes1000",
    )
    PriceUnit: MarketPriceUnit = Field(
        title="PriceUnit",
    )
    MarketSlotName: str = Field(
        title="MarketSlotName",
    )
    IrlTimeUtc: Optional[str] = Field(
        title="IrlTimeUtc",
        default=None,
    )
    MessageId: Optional[str] = Field(
        title="MessageId",
        default=None,
    )
    TypeName: Literal["latest.price"] = "latest.price"
    Version: str = "000"

    @validator("FromGNodeAlias")
    def _check_from_g_node_alias(cls, v: str) -> str:
        try:
            check_is_left_right_dot(v)
        except ValueError as e:
            raise ValueError(
                f"FromGNodeAlias failed LeftRightDot format validation: {e}"
            )
        return v

    @validator("FromGNodeInstanceId")
    def _check_from_g_node_instance_id(cls, v: str) -> str:
        try:
            check_is_uuid_canonical_textual(v)
        except ValueError as e:
            raise ValueError(
                f"FromGNodeInstanceId failed UuidCanonicalTextual format validation: {e}"
            )
        return v

    @validator("PriceUnit")
    def _check_price_unit(cls, v: MarketPriceUnit) -> MarketPriceUnit:
        return as_enum(v, MarketPriceUnit, MarketPriceUnit.USDPerMWh)

    @validator("MarketSlotName")
    def _check_market_slot_name(cls, v: str) -> str:
        try:
            check_is_market_slot_name_lrd_format(v)
        except ValueError as e:
            raise ValueError(
                f"MarketSlotName failed MarketSlotNameLrdFormat format validation: {e}"
            )
        return v

    @validator("IrlTimeUtc")
    def _check_irl_time_utc(cls, v: Optional[str]) -> Optional[str]:
        if v is None:
            return v
        try:
            check_is_iso_format(v)
        except ValueError as e:
            raise ValueError(f"IrlTimeUtc failed IsoFormat format validation: {e}")
        return v

    @validator("MessageId")
    def _check_message_id(cls, v: Optional[str]) -> Optional[str]:
        if v is None:
            return v
        try:
            check_is_uuid_canonical_textual(v)
        except ValueError as e:
            raise ValueError(
                f"MessageId failed UuidCanonicalTextual format validation: {e}"
            )
        return v

    def as_dict(self) -> Dict[str, Any]:
        d = self.dict()
        del d["PriceUnit"]
        PriceUnit = as_enum(self.PriceUnit, MarketPriceUnit, MarketPriceUnit.default())
        d["PriceUnitGtEnumSymbol"] = MarketPriceUnitMap.local_to_type(PriceUnit)
        if d["IrlTimeUtc"] is None:
            del d["IrlTimeUtc"]
        if d["MessageId"] is None:
            del d["MessageId"]
        return d

    def as_type(self) -> str:
        return json.dumps(self.as_dict())


class LatestPrice_Maker:
    type_name = "latest.price"
    version = "000"

    def __init__(
        self,
        from_g_node_alias: str,
        from_g_node_instance_id: str,
        price_times1000: int,
        price_unit: MarketPriceUnit,
        market_slot_name: str,
        irl_time_utc: Optional[str],
        message_id: Optional[str],
    ):
        self.tuple = LatestPrice(
            FromGNodeAlias=from_g_node_alias,
            FromGNodeInstanceId=from_g_node_instance_id,
            PriceTimes1000=price_times1000,
            PriceUnit=price_unit,
            MarketSlotName=market_slot_name,
            IrlTimeUtc=irl_time_utc,
            MessageId=message_id,
            #
        )

    @classmethod
    def tuple_to_type(cls, tuple: LatestPrice) -> str:
        """
        Given a Python class object, returns the serialized JSON type object
        """
        return tuple.as_type()

    @classmethod
    def type_to_tuple(cls, t: str) -> LatestPrice:
        """
        Given a serialized JSON type object, returns the Python class object
        """
        try:
            d = json.loads(t)
        except TypeError:
            raise SchemaError("Type must be string or bytes!")
        if not isinstance(d, dict):
            raise SchemaError(f"Deserializing {t} must result in dict!")
        return cls.dict_to_tuple(d)

    @classmethod
    def dict_to_tuple(cls, d: dict[str, Any]) -> LatestPrice:
        d2 = dict(d)
        if "FromGNodeAlias" not in d2.keys():
            raise SchemaError(f"dict {d2} missing FromGNodeAlias")
        if "FromGNodeInstanceId" not in d2.keys():
            raise SchemaError(f"dict {d2} missing FromGNodeInstanceId")
        if "PriceTimes1000" not in d2.keys():
            raise SchemaError(f"dict {d2} missing PriceTimes1000")
        if "PriceUnitGtEnumSymbol" not in d2.keys():
            raise SchemaError(f"dict {d2} missing PriceUnitGtEnumSymbol")
        if d2["PriceUnitGtEnumSymbol"] in MarketPriceUnit000SchemaEnum.symbols:
            d2["PriceUnit"] = MarketPriceUnitMap.type_to_local(
                d2["PriceUnitGtEnumSymbol"]
            )
        else:
            d2["PriceUnit"] = MarketPriceUnit.default()
        if "MarketSlotName" not in d2.keys():
            raise SchemaError(f"dict {d2} missing MarketSlotName")
        if "IrlTimeUtc" not in d2.keys():
            d2["IrlTimeUtc"] = None
        if "MessageId" not in d2.keys():
            d2["MessageId"] = None
        if "TypeName" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TypeName")

        return LatestPrice(
            FromGNodeAlias=d2["FromGNodeAlias"],
            FromGNodeInstanceId=d2["FromGNodeInstanceId"],
            PriceTimes1000=d2["PriceTimes1000"],
            PriceUnit=d2["PriceUnit"],
            MarketSlotName=d2["MarketSlotName"],
            IrlTimeUtc=d2["IrlTimeUtc"],
            MessageId=d2["MessageId"],
            TypeName=d2["TypeName"],
            Version="000",
        )



from enum import auto
from typing import List

from fastapi_utils.enums import StrEnum


class EmitterPumpFeedbackModel(StrEnum):
    """


    Choices and descriptions:

      * ConstantDeltaT:
      * ConstantGpm:
    """

    ConstantDeltaT = auto()
    ConstantGpm = auto()

    @classmethod
    def default(cls) -> "EmitterPumpFeedbackModel":
        """
        Returns default value ConstantDeltaT
        """
        return cls.ConstantDeltaT

    @classmethod
    def values(cls) -> List[str]:
        """
        Returns enum choices
        """
        return [elt.value for elt in cls]



from enum import auto
from typing import List

from fastapi_utils.enums import StrEnum


class LocalCommInterface(StrEnum):
    """
    Categorization of in-house comm mechanisms for SCADA

    Choices and descriptions:

      * Analog_4_20_mA:
      * RS232:
      * I2C:
      * Wifi:
      * SimRabbit:
      * Unknown:
      * Ethernet:
      * OneWire:
      * RS485:
    """

    ANALOG_4_20_MA = auto()
    RS232 = auto()
    I2C = auto()
    WIFI = auto()
    SIMRABBIT = auto()
    UNKNOWN = auto()
    ETHERNET = auto()
    ONEWIRE = auto()
    RS485 = auto()

    @classmethod
    def default(cls) -> "LocalCommInterface":
        """
        Returns default value UNKNOWN
        """
        return cls.UNKNOWN

    @classmethod
    def values(cls) -> List[str]:
        """
        Returns enum choices
        """
        return [elt.value for elt in cls]



""" Definition of a node in the Dijsktra graph for solving Forward Looking Optimizations"""
from abc import ABC
from typing import Optional


SIG_FIGS_FOR_OUTPUT = 6


class DNode(ABC):
    SIG_FIGS_FOR_OUTPUT = 6

    def __init__(
        self,
        ts_idx: int,
        store_idx: int,
        local_benefit: float = 0,
    ):
        self.ts_idx = ts_idx
        self.store_idx = store_idx
        self.local_benefit = local_benefit
        self.path_benefit: Optional[float] = None
        self.best_next_idx: Optional[int] = None
        self.path_edge_cost_only: Optional[float] = None

    @property
    def path_cost(self) -> Optional[float]:
        if self.path_benefit is None:
            return None
        return -self.path_benefit

    def __repr__(self) -> str:
        if self.local_benefit:
            local_benefit = round(self.local_benefit, self.SIG_FIGS_FOR_OUTPUT)
        else:
            local_benefit = None
        if self.path_benefit:
            path_benefit = round(self.path_benefit, self.SIG_FIGS_FOR_OUTPUT)
        else:
            path_benefit = None
        if self.path_edge_cost_only:
            path_edge_cost_only = round(
                self.path_edge_cost_only, self.SIG_FIGS_FOR_OUTPUT
            )
        else:
            path_edge_cost_only = None
        return (
            f"DNode => TsIdx: {self.ts_idx}, StoreIdx: {self.store_idx}, BestNextIdx: {self.best_next_idx}, "
            f"LocalBenefit: {local_benefit}, PathBenefit: {path_benefit}, PathEdgeCostOnly: {path_edge_cost_only}"
        )
